syntax = "proto2";

package issue270.a;

import "github.com/gogo/protobuf/gogoproto/gogo.proto";
import "github.com/gogo/protobuf/test/issue270/a/a2.proto";

option (gogoproto.populate_all) = true;

message A1 {
    optional A2 a2 = 1;
}
