// Copyright 2016 CNI authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package firewall

import (
	"fmt"
)

// FirewallNone is a firewall backend for environments where manipulating the
// system firewall is unsupported (for example, when running without root).
// Nolint applied to avoid firewall.FirewallNone name duplication notes.
//nolint
type FirewallNone struct{}

func newNoneBackend() (FirewallBackend, error) {
	return &FirewallNone{}, nil
}

// Add adds a rule to the system firewall.
// No action is taken and an error is unconditionally returned as this backend
// does not support manipulating the firewall.
func (f *FirewallNone) Add(conf *FirewallNetConf) error {
	return fmt.Errorf("cannot modify system firewall rules")
}

// Del deletes a rule from the system firewall.
// No action is taken and an error is unconditionally returned as this backend
// does not support manipulating the firewall.
func (f *FirewallNone) Del(conf *FirewallNetConf) error {
	return fmt.Errorf("cannot modify system firewall rules")
}
