/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.1
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

//  Status of a process running in a container
type ProcessStatus struct {

	ProcessId int32 `json:"ProcessId,omitempty"`

	Exited bool `json:"Exited,omitempty"`

	ExitCode int32 `json:"ExitCode,omitempty"`

	LastWaitResult int32 `json:"LastWaitResult,omitempty"`
}
